#Dans Terminal:
#jupyter lab
#python tp1.py


import matplotlib.pyplot as plt
import numpy as np

def orthonorm(y):
    (n,m)=np.shape(y)
    z=np.zeros((n,m))
    z[:,0]=y[:,0]/np.linalg.norm(y[:,0])
    for j in range(1,m):
        v=y[:,j]-np.dot(np.dot(z,z.T),y[:,j])
        z[:,j]=v/np.linalg.norm(v)
    return z




def qr(a):
    (n,m)=np.shape(a)
    a1=a.copy()
    r=min(n,m)
    h1=np.eye(n,n)
    for i in range(r-1):
        b1=a1.copy()
        x=a1[i:,i]
        x.shape=(n-i,1)
        q1,r1=np.linalg.qr(x)
        x[0]=x[0]+r1
        u=x/np.linalg.norm(x)
        h=np.eye(n,n)
        h[i:,i:]=np.eye(n-i,n-i)-2*np.dot(u,u.T)
        h1=np.dot(h,h1)
        a1=np.dot(h,b1)
    return h1.T,a1
    
    
def puiss_mat(a,m,eps):
    (l,n)=np.shape(a)
    x=np.random.rand(n,m)
    crit=1
    itmax=200
    i=0
    while (crit>eps) and (i<itmax):
        y=np.dot(a,x)
        nu=np.dot(x.T,y)
        x=orthonorm(y)
        crit=np.linalg.norm(np.triu(nu)-np.tril(nu))
        i=i+1
    return np.diag(nu),i
    
def puiss_mat1(a,m,eps):
    (l,n)=np.shape(a)
    x=np.random.rand(n,m)
    crit=1
    itmax=200
    i=0
    tabcrit=np.zeros(itmax)
    tabit=np.zeros(itmax)
    while (crit>eps) and (i<itmax):
        y=np.dot(a,x)
        nu=np.dot(x.T,y)
        x=orthonorm(y)
        #crit=np.linalg.norm(np.triu(nu)-np.tril(nu))
        v=np.diag(nu)
        crit=np.linalg.norm(v-np.diag(np.diag(100*0.9**(np.arange(1,m+1)))))
        tabit[i]=i+1
        tabcrit[i]=crit
        i=i+1
    return v,tabcrit[:i],tabit[:i]
    


def mat_puiss_(a,m):
    (l,c)=np.shape(a)
    b=np.eye(l,c)
    s=np.eye(l,c)
    for k in range(m):
        b=np.dot(b,a)/(k+1)
        s=s+b
    return s

def expm(a,eps):
    (l,c)=np.shape(a)
    b=np.eye(l,c)
    s=np.eye(l,c)
    crit=1
    k=0
    itmax=100
    while(crit>eps) and (k<itmax):
        s1=s.copy()
        b=np.dot(b,a)/(k+1)
        s=s+b
        crit=np.linalg.norm(s1-s)
        k=k+1
    return s,k





eps=1e-3

a1=np.array([[1,0.8,0.6],[0.8,0.2,0.8],[0,1.2,0.9]])
print('a1=',a1)

a2=np.array([[2,3,2],[1,2,3],[4,3,5.2]])
print('a2=',a2)

expa1,i=expm(a1,eps)
print('exponentielle de ',a1,'obtenue en ',i,'pas:',expa1)


expa2,i=expm(a2,eps)
print('exponentielle de ',a2,'obtenue en ',i,'pas:',expa2)

n=40
m=10

J=np.tril(np.ones((n,n)))
J=(J-J.T)/2
U,k=expm(J,1e-7)
print('Erreur sur U:')
print(np.linalg.norm(np.dot(U,U.T)-np.eye(n,n))+np.linalg.norm(np.dot(U.T,U)-np.eye(n,n)))


A=np.dot(np.dot(U,np.diag(100*0.9**(np.arange(1,n+1)))),U.T)
    
v,i=puiss_mat(A,m,eps)

#print('obtenu en ',i,'pas')
#print(v)

# print(np.diag(np.diag(100*0.9**(np.arange(1,m+1)))))
print(np.linalg.norm(v-np.diag(np.diag(100*0.9**(np.arange(1,m+1))))))

taberr=np.zeros(30)
for m in range(1,31):
    v,i=puiss_mat(A,m,eps)
    err=np.linalg.norm(v-np.diag(np.diag(100*0.9**(np.arange(1,m+1)))))
    taberr[m-1]=err
    
plt.plot(taberr[2:])
plt.show()
    
v,i=puiss_mat(A,m,eps)
print(np.linalg.norm(v-np.diag(np.diag(100*0.9**(np.arange(1,m+1))))))

m=10
v,taberr,tabit=puiss_mat1(A,m,eps)
plt.plot(tabit[1:],np.log(taberr[1:]))
plt.show()

P=np.polyfit(tabit[1:],np.log(taberr[1:]),1)

rho_exp=np.exp(round(P[0],2))
print('rho experimental pour m=',m,':',rho_exp)


m=20
v,taberr,tabit=puiss_mat1(A,m,eps)
plt.plot(tabit[1:],np.log(taberr[1:]))
plt.show()

P=np.polyfit(tabit[1:],np.log(taberr[1:]),1)

rho_exp=np.exp(round(P[0],2))
print('rho experimental pour m=',m,':',rho_exp)

tabrho=np.zeros(20)

for m in range(1,21):
    v,taberr,tabit=puiss_mat1(A,m,eps)
    plt.plot(tabit[1:],np.log(taberr[1:]))
    P=np.polyfit(tabit[1:],np.log(taberr[1:]),1)
    tabrho[m-1]=np.exp(round(P[0],2))
plt.show()

plt.plot(tabrho)
plt.title('Graphe des erreurs en fonction du nombre de valeurs calculees')
plt.show()
    

        
    
